package server

import (
	"context"
	"encoding/json"
	"log" // TODO: Use a request scoped logger.
	"net/http"
	"reflect"

	"github.com/deref/exo/jsonutil"
)

type MethodHandler struct {
	f   reflect.Value
	in  reflect.Type
	out reflect.Type
}

func NewMethodHandler(f interface{}) *MethodHandler {
	fv := reflect.ValueOf(f)
	if fv.Kind() != reflect.Func {
		panic("expected function, got " + fv.Kind().String())
	}
	typ := fv.Type()
	if typ.NumIn() != 2 {
		panic("wrong number of parameters. " + expectedSignature)
	}
	if typ.NumOut() != 2 {
		panic("wrong number of results. " + expectedSignature)
	}
	if typ.In(0) != contextType {
		panic("first parameter must be a context. " + expectedSignature)
	}
	in := typ.In(1)
	if in.Kind() != reflect.Ptr {
		panic("first parameter must be a pointer. " + expectedSignature)
	}
	out := typ.Out(0)
	if out.Kind() != reflect.Ptr {
		panic("first result must be a pointer. " + expectedSignature)
	}
	if typ.Out(1) != errorType {
		panic("second result must be an error. " + expectedSignature)
	}
	return &MethodHandler{
		f:   fv,
		in:  in.Elem(),
		out: out,
	}
}

const expectedSignature = "expected signature: func (ctx context.Context, input *YourInput) (*YourOutput, error)"

var contextType = reflect.TypeOf((*context.Context)(nil)).Elem()
var errorType = reflect.TypeOf((*error)(nil)).Elem()

func (handler *MethodHandler) ServeHTTP(w http.ResponseWriter, req *http.Request) {
	// TODO: Figure out the "right" thing to do for cors.
	w.Header().Set("Access-Control-Allow-Origin", "*")
	w.Header().Set("Access-Control-Allow-Methods", "*")
	w.Header().Set("Access-Control-Allow-Headers", "*")

	if req.Method == "OPTIONS" {
		w.WriteHeader(http.StatusOK)
		return
	}

	// TODO: Check content-type, accepts, etc.
	// TODO: Include a Request ID in logs?
	// TODO: Differentiate 400s from 500s, internal vs external errors, etc.
	input := reflect.New(handler.in)
	if req.Body != nil {
		if err := jsonutil.UnmarshalReader(req.Body, input.Interface()); err != nil {
			w.WriteHeader(http.StatusBadRequest)
			log.Printf("error parsing request: %v", err)
			w.Write([]byte("error parsing json\n"))
			return
		}
	}
	results := handler.f.Call([]reflect.Value{
		reflect.ValueOf(req.Context()),
		input,
	})
	output := results[0].Interface()
	errv := results[1].Interface()
	if errv != nil {
		err := errv.(error)
		w.WriteHeader(http.StatusInternalServerError)
		log.Printf("error processing request: %v", err)
		w.Write([]byte("internal server error\n"))
		return
	}
	w.Header().Set("Content-Type", "application/json")

	w.WriteHeader(http.StatusOK)
	enc := json.NewEncoder(w)
	if err := enc.Encode(output); err != nil {
		log.Printf("error encoding response: %v", err)
	}
}

type MuxBuilder struct {
	prefix string
	mux    *http.ServeMux
}

func NewMuxBuilder(prefix string) *MuxBuilder {
	return &MuxBuilder{
		prefix: prefix,
		mux:    http.NewServeMux(),
	}
}

func (b *MuxBuilder) Mux() *http.ServeMux {
	return b.mux
}

func (b *MuxBuilder) AddMethod(name string, f interface{}) {
	b.mux.Handle(b.prefix+name, NewMethodHandler(f))
}
